
/* Most are taken directly from src/internal.h or src/djgpp/interndj.h */

#define POLYGON_FIX_SHIFT     18


/* bitfield specifying which polygon attributes need interpolating */
#define INTERP_FLAT           1
#define INTERP_1COL           2
#define INTERP_3COL           4
#define INTERP_FIX_UV         8
#define INTERP_Z              16
#define INTERP_FLOAT_UV       32
#define OPT_FLOAT_UV_TO_FIX   64
#define COLOR_TO_RGB          128
#define INTERP_ZBUF           256
#define INTERP_TRANS          512
#define INTERP_BLEND          1024
#define INTERP_THRU           2048


/* information for polygon scanline fillers */
typedef struct POLYGON_SEGMENT
{
   fixed u, v, du, dv;              /* fixed point u/v coordinates */
   fixed c, dc;                     /* single color gouraud shade values */
   fixed r, g, b, dr, dg, db;       /* RGB gouraud shade values */
   float z, dz;                     /* polygon depth (1/z) */
   float fu, fv, dfu, dfv;          /* floating point u/v coordinates */
   unsigned char *texture;          /* the texture map */
   int umask, vmask, vshift;        /* texture map size information */
   int seg;                         /* destination bitmap selector */
} POLYGON_SEGMENT;


/* prototype for the scanline filler functions */
typedef void (*SCANLINE_FILLER)(unsigned long addr, int w, POLYGON_SEGMENT *info);


typedef struct POLYGON_EDGE      /* an active edge */
{
   int top;                      /* top y position */
   int bottom;                   /* bottom y position */
   fixed x, dx;                  /* fixed point x position and gradient */
   fixed w;                      /* width of line segment */
   POLYGON_SEGMENT dat;          /* texture/gouraud information */
   struct POLYGON_EDGE *prev;    /* doubly linked list */
   struct POLYGON_EDGE *next;
   /* from here on add-ons to the Allegro structure */
   struct POLYGON_INFO *poly;    /* father polygon */
} POLYGON_EDGE;


typedef struct POLYGON_INFO {               /* a polygon waiting rendering */
   struct POLYGON_INFO *next, *prev;        /* double linked list */
   int             inside;                  /* flag for "scanlining" */
   int             flags;                   /* INTERP_* flags */
   int             color;                   /* vtx[0]->c */
   float           a, b, c;                 /* plane's coefficients -a/d, -b/d, -c/d */
   int             alpha;                   /* blender alpha */
   int             b15, b16, b24, b32;      /* blender colors */
   COLOR_MAP       *cmap;                   /* trans color map */
   SCANLINE_FILLER drawer, alt_drawer;      /* scanline drawing functions */
   POLYGON_EDGE    *left_edge, *right_edge; /* true edges used in interpolation */
   POLYGON_SEGMENT info;                    /* base information for scanline functions */
} POLYGON_INFO;


extern SCANLINE_FILLER _optim_alternative_drawer;

POLYGON_EDGE *_add_edge(POLYGON_EDGE *list, POLYGON_EDGE *edge, int sort_by_x);
POLYGON_EDGE *_remove_edge(POLYGON_EDGE *list, POLYGON_EDGE *edge);
void _fill_3d_edge_structure(POLYGON_EDGE *edge, V3D *v1, V3D *v2, int flags, BITMAP *bmp);
void _fill_3d_edge_structure_f(POLYGON_EDGE *edge, V3D_f *v1, V3D_f *v2, int flags, BITMAP *bmp);
SCANLINE_FILLER _get_scanline_filler(int type, int *flags, POLYGON_SEGMENT *info, BITMAP *texture, BITMAP *bmp);
void _clip_polygon_segment(POLYGON_SEGMENT *info, int gap, int flags);

extern int _drawing_mode;
extern unsigned long _zbuf_addr;

/* polygon scanline filler functions */
void _poly_scanline_dummy(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_gcol8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_grgb8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb8x(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb15x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit15x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit15x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit15x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit15x(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_ptex_lit15d(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit15d(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb16x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit16x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit16x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit16x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit16x(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_ptex_lit16d(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit16d(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb24x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit24x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit24x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit24x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit24x(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_ptex_lit24d(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit24d(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_grgb32x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_lit32x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_lit32x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_atex_mask_lit32x(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit32x(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_scanline_ptex_lit32d(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_scanline_ptex_mask_lit32d(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_zbuf_flat8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_gcol8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_grgb8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_lit8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_trans8(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_zbuf_gcol15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_grgb15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_lit15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_trans15(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_zbuf_flat16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_grgb16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_lit16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_trans16(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_zbuf_flat24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_grgb24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_lit24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_trans24(unsigned long addr, int w, POLYGON_SEGMENT *info);

void _poly_zbuf_flat32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_grgb32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_lit32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_atex_mask_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);
void _poly_zbuf_ptex_mask_trans32(unsigned long addr, int w, POLYGON_SEGMENT *info);

/* some Allegro functions need a block of scratch memory */
extern void *_scratch_mem;
extern int _scratch_mem_size;

__INLINE__ void _grow_scratch_mem(int size)
{
   if (size > _scratch_mem_size) {
      size = (size+1023) & 0xFFFFFC00;
      _scratch_mem = realloc(_scratch_mem, size);
      _scratch_mem_size = size;
   }
}

